#include "scoring.hh"
#include "rna_data.hh"
#include "sequence.hh"

#include <math.h>

void Scoring::precompute_sigma() {
    int lenA = bpsA->get_rnadata()->get_sequence().length();
    int lenB = bpsB->get_rnadata()->get_sequence().length();
    
    sigma_tab.resize(lenA+1,lenB+1);

    for (int i=1; i<=lenA; ++i) {
	for (int j=1; j<=lenB; ++j) {
	    sigma_tab(i,j) = sigma_(i,j);
	}
    }
}

/**
   returns similarity of two alignment columns
 */
score_t
Scoring::sigma_(int ia, int ib) const {
  
    if (params->mea_scoring) {
	
	// for our mea alignment, we score basematchs
	// by the sum of
	//   the edge probability
	//   and a term for structural accuracy
	
	return
	    static_cast<score_t>
	    (
		score_resolution *
		( params->match_probs->prob(ia,ib)
		  +
		  (params->alpha_factor/100.0)
		  * (
		      bpsA->get_rnadata()->prob_unpaired(ia)
		      +
		      bpsB->get_rnadata()->prob_unpaired(ib)
		      )
		    )
		);
    } else {
	// compute average score for aligning the two alignment columns

	const Sequence &seqA = bpsA->get_rnadata()->get_sequence();
	const Sequence &seqB = bpsB->get_rnadata()->get_sequence();
	
	const AliColumn &colA=seqA[ia];
	const AliColumn &colB=seqB[ib];
	
	score_t score=0;
	
	// the sum of pairs is quite ad hoc
	// e.g. matching - and - counts as match ...
	// N is a wildcard, matchs with N don't count
	//
	for (size_t i=0; i<colA.size(); i++) {
	    for (size_t j=0; j<colB.size(); j++) {
		// if we have a ribosum matrix, use it!
		// !!! if ribosum and characters are not in the ribosum matrix
		// we fall back to match/mismatch scoring !!! 
		//
		if (params->ribosum 
		    && params->ribosum->alphabet().in(colA[i])
		    && params->ribosum->alphabet().in(colB[j])) 
		{
		    score +=
			(score_t)(100.0 * params->ribosum->basematch_score(colA[i],colB[j]));
		} else {
		    if (colA[i]!='N' && colB[j]!='N') {
			score +=
			    (colA[i]==colB[j])
			    ?params->basematch
			    :params->basemismatch;
		    }
		}
	    }
	}
	
	return  (score_t)(score / (int)(colA.size()*colB.size())) ;
    }
}

void
Scoring::precompute_weights(const BasePairs &bps,
			    std::vector<score_t> &weights,
			    std::vector<score_t> &stack_weights) {
    size_t s=bps.size();
    size_t len=bps.get_rnadata()->get_sequence().length();
    weights.resize(s);
    if (params->stacking) stack_weights.resize(s);
    for (size_type i=0; i<s; i++) {
	const Arc &a=bps.arc(i);
	
	double p = bps.get_rnadata()->get_arc_prob(a.left(),a.right());
	weights[i] = probToWeight(p,len);
	
	if (params->stacking) {
	    double stack_p = bps.get_rnadata()->get_arc_stack_prob(a.left(),a.right());
	    stack_weights[i] = probToWeight(stack_p,len);
	}
    }
}

void
Scoring::precompute_weights() {
    //score_t weight = 
    //score_t cond_weight = probToWeight(cond_prob);
    
    precompute_weights(*bpsA,weightsA,stack_weightsA);
    precompute_weights(*bpsB,weightsB,stack_weightsB);
}

score_t
Scoring::probToWeight(double p, size_t len) const
{
    if (params->mea_scoring) {
	return static_cast<score_t>(score_resolution*p);
    } else {
	double pe=prob_exp(len);
	
	return
	    static_cast<score_t> (
		round(params->struct_weight*
		      (1-log(p)/log(pe)))
		);
        /*
	  Score soll 0 sein für p=pe und struct_weight für p=1
	  
	  ==> Normalisierung log(p/pe) / log(1/pe) 
	  //                  [=1-log(p)/log(pe)]
	  
	*/
    }
}

double Scoring::prob_exp(size_t len) const {
    if (params->exp_prob >= 0) {
	return  params->exp_prob;
    } else {
	return prob_exp_f(len);
    }
}


// here we implement a special treatment of gap cost in multiple alignment
// gap cost is varied position specific in order to favor gapping columns
// that are already gapped.
// This can be refined further!
void
Scoring::precompute_gapcost() {

    const Sequence &seqA = bpsA->get_rnadata()->get_sequence();
    const Sequence &seqB = bpsB->get_rnadata()->get_sequence();
	
    size_type lenA = seqA.length();
    size_type lenB = seqB.length();
    
    // resize and create tables
    gapcost_tabA.resize(lenA+1);
    gapcost_tabB.resize(lenB+1);
    std::vector<float> gapfreqA(lenA+1,0);
    std::vector<float> gapfreqB(lenB+1,0);
    
    // determine gap frequencies for each column in A and B

    for (size_type i=1; i<lenA+1; i++) {
	const AliColumn &colA=seqA[i];
	for (size_type k=0; k<colA.size(); k++) {
	    gapfreqA[i] += (colA[k]=='-')?1:0;
	}
	gapfreqA[i] /= colA.size();
    }

    for (size_type i=1; i<lenB+1; i++) {
	const AliColumn &colB=seqB[i];
	for (size_type k=0; k<colB.size(); k++) {
	    gapfreqB[i] += (colB[k]=='-')?1:0;
	}
	gapfreqB[i] /= colB.size();
    }
    // gap freqs determined
    
    // ----------------------------------------
    // compute position specific gap cost
    
    for (size_type i=1; i<lenA+1; i++) {
	gapcost_tabA[i] = (score_t) ((1-gapfreqA[i]) * params->indel);
    }
    
    for (size_type i=1; i<lenB+1; i++) {
	gapcost_tabB[i] = (score_t) ((1-gapfreqB[i]) * params->indel);
    }
    
}


double
Scoring::ribosum_arcmatch_prob(const Arc &arcA, const Arc &arcB) const {
    assert(params->ribosum != 0);
    // compute average ribosum score
    
    RibosumFreq *ribosum = static_cast<RibosumFreq *>(params->ribosum);
    
    const Sequence &seqA = bpsA->get_rnadata()->get_sequence();
    const Sequence &seqB = bpsB->get_rnadata()->get_sequence();
	
    const size_type rowsA = seqA.get_rows();
    const size_type rowsB = seqB.get_rows();
    
    double score=0;
    int gapless_combinations=0;
	
    // compute geometric mean
    for(size_type i=0; i<rowsA; i++) { // run through all combinations of rows in A and B
	for(size_type j=0; j<rowsB; j++) { 
	    // wie sollen gaps behandelt werden?
	    // current solution: ignore gap entries
		
	    if (seqA[arcA.left()][i]!='-'
		&& seqA[arcA.right()][i]!='-'
		&& seqB[arcB.left()][j]!='-'
		&& seqB[arcB.right()][j]!='-') {
		    
		gapless_combinations++;
		    
		score+= 
		    log(
			ribosum->arcmatch_prob( seqA[arcA.left()][i],seqA[arcA.right()][i],
						seqB[arcB.left()][j],seqB[arcB.right()][j] )
			/
			( ribosum->basematch_prob(seqA[arcA.left()][i],seqB[arcB.left()][j])
			  *
			  ribosum->basematch_prob(seqA[arcA.right()][i],seqB[arcB.right()][j]))
			);
	    }
	}
    }
    
    return exp(score / gapless_combinations);
}

score_t
Scoring::ribosum_arcmatch_score(const Arc &arcA, const Arc &arcB) const {
    assert(params->ribosum != 0);
    
    // compute average ribosum score
    
    RibosumFreq *ribosum = static_cast<RibosumFreq *>(params->ribosum);
    
    const Sequence &seqA = bpsA->get_rnadata()->get_sequence();
    const Sequence &seqB = bpsB->get_rnadata()->get_sequence();
	
    const size_type rowsA = seqA.get_rows();
    const size_type rowsB = seqB.get_rows();
	
    double score=0;
    int gapless_combinations=0;
    
    for(size_type i=0; i<rowsA; i++) { // run through all combinations of rows in A and B
	for(size_type j=0; j<rowsB; j++) { 
	    // wie sollen gaps behandelt werden?
	    // current solution: ignore gap entries
	    
	    if (seqA[arcA.left()][i]!='-'
		&& seqA[arcA.right()][i]!='-'
		&& seqB[arcB.left()][j]!='-'
		&& seqB[arcB.right()][j]!='-') {
		    
		gapless_combinations++;
		
		score+= 
		    log(
			ribosum->arcmatch_prob( seqA[arcA.left()][i],seqA[arcA.right()][i],
						seqB[arcB.left()][j],seqB[arcB.right()][j] )
			/
			( ribosum->basepair_prob(seqA[arcA.left()][i],seqA[arcA.right()][i])
			  *
			  ribosum->basepair_prob(seqB[arcB.left()][j],seqB[arcB.right()][j]))
			);
	    }
	}
    }
	
    return (score_t) (100.0 * score / gapless_combinations);
}
